// WakeUpMan Arduino Code v0.1.1
// Written By David de Gelder, March 2024

// ## Libraries

#include <LiquidCrystal_I2C.h>
#include <ezButton.h>

// ## Time Declarations

typedef struct Time {
public:
  int hours = 0;
  int minutes = 0;
  int seconds = 0;
  int elapsed_mills = 0;
  int current_mills = 0;
} Time;

Time t;     // Current Time, automatically changing
Time alarm; // Time to sound alarm, static unless manually changed

// Used when setting a time
Time tempTime;
Time* editing_time;
int* editing_var;


// ## UI declarations
// This menu system allows for nested menus with simple 3-button navigation between them. 

typedef struct MenuItem {
public:
  String name; // title of this menu item, displayed on the first line.
  int length; // amounts of options
  String options[10]; // 
} MenuItem;

MenuItem mainMenu = { "Menu", 4,
                      "Set Time",
                      "Set Alarm",
                      "Demo Alarm",
                      "Close" };


// Example of possible nested menu, currently unused:
/*
MenuItem alarmMenu = { "Set Alarm", 6,
                       "Everyday",
                       "Tomorrow",
                       "Workweek",
                       "Weekend",
                       "Per Day",
                       "Back" };
*/

MenuItem currentMenu = mainMenu;
int currentOption = 0; // index to the MenuItem.options array

// Possible states of the display
enum State { IN_MENU,
             INPUT_TIME,
             SHOW_TIME,
             DISPLAY_OFF, 
             SOUND_ALARM};

int state = SHOW_TIME;

// Button events
bool buttonLeft, buttonRight, buttonEnter, anyButtonDown, anyButtonPressed;


// ## Hardware Connections ##

ezButton button1(2), button2(3), button3(4);
LiquidCrystal_I2C lcd(0x3F, 16, 2);
int relayPin = 12;


// Runs once at device startup/reset
void setup() {
  // Open Serial
  Serial.begin(9600);  // For debugging

  // Set default time and alarm
  t = setTime(t, "07:30");
  alarm = setTime(alarm, "07:30");

  // Ready buttons
  button1.setDebounceTime(50);
  button2.setDebounceTime(50);
  button3.setDebounceTime(50);

  // Ready LCD
  lcd.init();
  lcd.backlight();
  lcd.clear();

  // Ready Relay
  pinMode(relayPin, OUTPUT);
}

// Core system loop
void loop() {
  // Update current time
  updateTime();

  // Sound alarm when needed
  if (checkAlarm(&alarm)) {
    soundAlarm();
  }
  
  // Check for button presses
  processInput();

  // Handle input and display to LCD
  display();
}

// ## Display Functions

// Display correct screen to LCD based on current UI state.
void display() {
  switch (state) {
    case IN_MENU:
      navigateMenu();
      break;
    case INPUT_TIME:
      inputTime();
      break;
    case SHOW_TIME:
      displayClock();
      break;
    case SOUND_ALARM:
      displayAlarm();
      break;
  }
}

void navigateMenu() {
  // Scroll Left/right through options
  if (buttonLeft) {
    currentOption -= 1;
  }
  if (buttonRight) {
    currentOption += 1;
  }

  // Cycles options when reaching the end/beginning
  currentOption = mod(currentOption, currentMenu.length);

  // Only display when changes are made, to avoid flashing
  if (anyButtonDown) {
    displayMenu();
  }

  // Enter submenu
  if (buttonEnter) {
    menuEnter(currentMenu.options[currentOption]);
  }
}

void displayMenu() {
  // Print menu title to first line
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(currentMenu.name);

  // Print navigation arrows to 2nd line
  lcd.setCursor(0, 1);
  lcd.print("<              >");

  // Print current selected options to 2nd line
  lcd.setCursor(2, 1);
  lcd.print(currentMenu.options[currentOption]);
}

void menuEnter(String option) {
  lcd.clear();
  currentOption = 0; // Select first option again

  // Navigate through the menu based on the selected option.

  if (option == "Back") {
    currentMenu = mainMenu;
  }

  if (option == "Set Alarm") {
    state = INPUT_TIME;
    startInputTime(&alarm);
  }

  if (option == "Set Time") {
    state = INPUT_TIME;
    startInputTime(&t);
  }

  if (option == "Close") {
    lcd.clear();
    state = SHOW_TIME;
  }

  if (option == "Demo Alarm"){ // Set time to 5 seconds before the alarm goes off
    t = setTime(t, formatTime(alarm, "%02d:%02d"));
    t.minutes -= 1;
    t.seconds = 55;
    menuEnter("Close");
  }
}

void startInputTime(Time* edit) {
  editing_time = edit; //pointer to Time variable to edit
  tempTime = *editing_time;
  editing_var = &tempTime.hours; // Edit hours first

  lcd.clear();

  // First line
  lcd.setCursor(0, 0);
  lcd.print("Set Time:");

  // Second line
  lcd.setCursor(5, 1);
  lcd.print(formatTime(tempTime, "%02d:%02d"));

  // Show cursor where user can edit the variable
  lcd.setCursor(6, 1);
  lcd.cursor();
}

void inputTime() {
  // Increase/Decrease current variable
  if (buttonLeft) {
    *editing_var -= 1;
  }

  if (buttonRight) {
    *editing_var += 1;
  }

  // Keep the numbers within possible range
  tempTime.hours = mod(tempTime.hours, 24);
  tempTime.minutes = mod(tempTime.minutes, 60);

  // Update display on changes
  if (anyButtonPressed) {
    lcd.setCursor(5, 1);
    lcd.print(formatTime(tempTime, "%02d:%02d"));
  }

  // When User presses enter
  if (buttonEnter) {
    // first time, move from hours to minutes
    if (editing_var == &tempTime.hours) {
      editing_var = &tempTime.minutes;
    // second time, apply changes
    } else if (editing_var == &tempTime.minutes) {
      *editing_time = setTime(*editing_time, formatTime(tempTime, "%02d:%02d"));

      // Go back to showing time
      state = SHOW_TIME;
      lcd.noCursor();
      lcd.clear();
    }
  }

  // Move cursor to currently edited variable 
  if (editing_var == &tempTime.hours) { lcd.setCursor(6, 1); }
  if (editing_var == &tempTime.minutes) { lcd.setCursor(9, 1); }
}

void displayClock() {
  // First Line: Current Time
  lcd.setCursor(0,0);
  lcd.print("===");

  lcd.setCursor(13,0);
  lcd.print("===");

  lcd.setCursor(4, 0);
  lcd.print(formatTime(t, "%02d:%02d:%02d"));

  // Second Line: Message based on time
  lcd.setCursor(0, 1);
  lcd.print(getMessage(t));

  // Turn off screen
  if (buttonLeft){
    lcd.noBacklight();
  }

  // Turn on Screen
  if (buttonRight){
    lcd.backlight();
  }

  // Go into Menu
  if (buttonEnter) {
    lcd.backlight();
    state = IN_MENU;
    menuEnter("Back");
  }
}

void displayAlarm() {
  // First line: Current Time
  lcd.setCursor(0,0);
  lcd.print("===");

  lcd.setCursor(13,0);
  lcd.print("===");

  lcd.setCursor(4, 0);
  lcd.print(formatTime(t, "%02d:%02d:%02d"));

  // Second line: WAKE UP!
  lcd.setCursor(0, 1);
  lcd.print("--Wake Up Man!--");

  // Turn off alarm when any button pressed
  if (anyButtonPressed)
  {
    state = SHOW_TIME;
    digitalWrite(relayPin, LOW);
  }

}

// ## Time Functions

void updateTime() {
  bool queuePrintTime = false;

  // Get elapsed mills since last update
  t.elapsed_mills = millis() - t.current_mills;

  if (t.elapsed_mills >= 1000) {
    // Extra logic to handle potential delays
    t.seconds += floor(t.elapsed_mills / 1000);
    t.current_mills = millis() - t.elapsed_mills % 1000;

    // queue printTime() after handling minutes & hours
    queuePrintTime = true;
  }

  if (t.seconds >= 60) {
    t.seconds -= 60;
    t.minutes++;
  }

  if (t.minutes == 60) {
    t.minutes = 0;
    t.hours++;
  }

  if (t.hours == 24) {
    t.hours = 0;
    //next day
  }

  if (queuePrintTime) { printTime(t); }
}

Time setTime(Time oldTime, String string) {
  // Set a time using a string formatted HH:MM:SS
  Time newTime;
  newTime = oldTime;

  newTime.hours = string.substring(0, 2).toInt();
  newTime.minutes = string.substring(3, 5).toInt();
  newTime.seconds = 0;

  return newTime;
}

String formatTime(Time time, const char* format) {
  char buffer[10];
  sprintf(buffer, format, time.hours, time.minutes, time.seconds);
  return buffer;
}

void printTime(Time time) {
  Serial.println(formatTime(t,"%02d:%02d:%02d"));
}

bool checkAlarm(Time* a) { 
  return t.hours == a->hours && t.minutes == a->minutes && t.seconds == 0;
}

void soundAlarm() {
  // Activate Relay
  digitalWrite(relayPin, HIGH);
  state = SOUND_ALARM;
}

String getMessage(Time time){
  // Custom messages based on time of day

  if (time.hours >= 6 & time.hours < 12){
    return "Good Morning! :)";
  }

  if (time.hours >= 12 & time.hours < 18){
    return "Have a nice day!";
  }

  if (time.hours >= 18 & time.hours < 22){
    return "Good Evening!";
  }

   if (time.hours >= 22 | time.hours < 6){
    return "Good Night! ZzzZzz";
  }


}

// ## Utility Functions

void processInput() {
  button1.loop();
  button2.loop();
  button3.loop();

  buttonLeft = button1.isPressed();
  buttonRight = button2.isPressed();
  buttonEnter = button3.isPressed();

  anyButtonDown = ((button1.getStateRaw() == LOW) | (button2.getStateRaw() == LOW) | (button3.getStateRaw() == LOW));
  anyButtonPressed = (buttonLeft | buttonRight | buttonEnter);
}

int mod(int a, int b) {
  // Inbuilt % operator doesn't handle negative numbers nicely, this does.
  return (a % b + b) % b; // don't ask me why this works
}
